/* Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of NVIDIA CORPORATION nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
#define WINDOWS_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#pragma warning(disable : 4819)
#endif

#include <Exceptions.h>
#include <ImageIO.h>
#include <string.h>
#include <fstream>
#include <iostream>
#include <cuda_runtime.h>
#include <npp.h>
#include <helper_cuda.h>
#include <helper_string.h>



int main(int argc, char** argv) {
	cudaError_t cudaError;
	NppStatus nppStatus;
	NppStreamContext nppStreamCtx;
	nppStreamCtx.hStream = 0; // The NULL stream by default, set this to whatever your stream ID is if not the NULL stream.

	cudaError = cudaGetDevice(&nppStreamCtx.nCudaDeviceId);
	if (cudaError != cudaSuccess)
	{
		printf("CUDA error: no devices supporting CUDA.\n");
		return NPP_NOT_SUFFICIENT_COMPUTE_CAPABILITY;
	}

	const NppLibraryVersion *libVer = nppGetLibVersion();

	printf("NPP Library Version %d.%d.%d\n", libVer->major, libVer->minor, libVer->build);

	int driverVersion, runtimeVersion;
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);

	printf("CUDA Driver  Version: %d.%d\n", driverVersion / 1000, (driverVersion % 100) / 10);
	printf("CUDA Runtime Version: %d.%d\n\n", runtimeVersion / 1000, (runtimeVersion % 100) / 10);

	cudaError = cudaDeviceGetAttribute(&nppStreamCtx.nCudaDevAttrComputeCapabilityMajor,
		cudaDevAttrComputeCapabilityMajor,
		nppStreamCtx.nCudaDeviceId);
	if (cudaError != cudaSuccess)
		return NPP_NOT_SUFFICIENT_COMPUTE_CAPABILITY;

	cudaError = cudaDeviceGetAttribute(&nppStreamCtx.nCudaDevAttrComputeCapabilityMinor,
		cudaDevAttrComputeCapabilityMinor,
		nppStreamCtx.nCudaDeviceId);
	if (cudaError != cudaSuccess)
		return NPP_NOT_SUFFICIENT_COMPUTE_CAPABILITY;

	cudaError = cudaStreamGetFlags(nppStreamCtx.hStream, &nppStreamCtx.nStreamFlags);

	cudaDeviceProp oDeviceProperties;

	cudaError = cudaGetDeviceProperties(&oDeviceProperties, nppStreamCtx.nCudaDeviceId);

	nppStreamCtx.nMultiProcessorCount = oDeviceProperties.multiProcessorCount;
	nppStreamCtx.nMaxThreadsPerMultiProcessor = oDeviceProperties.maxThreadsPerMultiProcessor;
	nppStreamCtx.nMaxThreadsPerBlock = oDeviceProperties.maxThreadsPerBlock;
	nppStreamCtx.nSharedMemPerBlock = oDeviceProperties.sharedMemPerBlock;

	/* Dummy data */

	const int input_size = 20;
	const int template_size = 20;
	const int output_size = 63;
	NppiSize srcRoiSize = {5,4};
	NppiSize tplRoiSize = {5,4};
	NppiSize dstRoiSize = {9,7};
	int srcStep = 5*sizeof(Npp32f);
	int tplStep = 5*sizeof(Npp32f);
	int dstStep = 9*sizeof(Npp32f);
	int input_size_in_bytes = input_size * sizeof(float);
	int output_size_in_bytes = output_size * sizeof(float);
	int template_size_in_bytes = template_size * sizeof(float);

	int host_input[input_size] = { 2, 2, 2, 2, 2, 
								   2, 2, 2, 2, 2,
								   2, 2, 2, 2, 2,
								   2, 2, 2, 2, 2};
	int host_template[template_size] = { 1, 1, 1, 1, 1,
										 1, 1, 1, 1, 1,
										 1, 1, 1, 1, 1,
										 1, 1, 1, 1, 1};
	float host_output[output_size];

	Npp8u *dev_input, *dev_template;
	cudaError = cudaMalloc(&dev_input, input_size_in_bytes);
	if (cudaError != cudaSuccess)
		return NPP_MEMORY_ALLOCATION_ERR;
	cudaError = cudaMalloc(&dev_template, template_size_in_bytes);
	if (cudaError != cudaSuccess)
		return NPP_MEMORY_ALLOCATION_ERR;

	Npp32f *dev_output;
	cudaError = cudaMalloc(&dev_output, output_size_in_bytes);
	if (cudaError != cudaSuccess)
		return NPP_MEMORY_ALLOCATION_ERR;

	//Copy data to device
	cudaMemcpy(dev_input, host_input, input_size_in_bytes, cudaMemcpyHostToDevice);
	//Copy template to device
	cudaMemcpy(dev_template, host_template, template_size_in_bytes, cudaMemcpyHostToDevice);

	// calling nppiSqrDistanceValid_Norm_8u32f_C1R_Ctx routine 
	nppStatus = nppiSqrDistanceValid_Norm_8u32f_C1R_Ctx(dev_input, srcStep, srcRoiSize, dev_template, tplStep, tplRoiSize, dev_output, dstStep, nppStreamCtx);

	if (nppStatus != NPP_SUCCESS)
	{
		printf("nppiSqrDistanceValid_Norm_8u32f_C1R failed.\n");
		return -1;
	}
	//Copy result to host
	cudaMemcpy(host_output, dev_output, output_size_in_bytes, cudaMemcpyDeviceToHost);

	for (size_t i = 0; i < 1; i++)
		std::cout << host_output[i] << ' ';
}

